import datetime
import json
import re

import isodate

# utility functions


def camelize(string, uppercase_first_letter=True):
    # copied from the inflection module
    if uppercase_first_letter:
        return re.sub(r"(?:^|_)(.)", lambda m: m.group(1).upper(), string)
    else:
        return string[0].lower() + camelize(string)[1:]

# base vocabulary


def as_uri(text):
    return 'https://www.w3.org/ns/activitystreams#' + text


def as_uri_prop(text):
    return as_uri(camelize(text, False))


class _Base:
    TYPE_PROPERTIES = []
    CONTEXT = 'https://www.w3.org/ns/activitystreams'

    def __init__(self, **kwargs):
        self._props = {
            '@type': self.get_type()
        }
        if 'id' in kwargs:
            value = kwargs.pop('id')
            if value:
                self._props['@id'] = value

        for k, v in kwargs.items():
            if k not in self.TYPE_PROPERTIES:
                raise TypeError('Unrecognized property {}'.format(k))
            k = as_uri_prop(k)
            self._props[k] = v

    @classmethod
    def get_type(cls):
        return as_uri(cls.__name__)

    def __str__(self):
        return json.dumps(self.to_dict(), indent=2)

    def to_dict(self):
        obj = self.part_to_dict()
        obj['@context'] = self.CONTEXT
        return obj

    def _coerce_prop(self, key, val):
        if hasattr(val, 'part_to_dict'):
            val = val.part_to_dict()
        return PROPERTIES[key](val, self.CONTEXT)

    def _props_to_dict(self):
        for key, value in self._props.items():
            if key not in PROPERTIES:
                raise TypeError('Unrecognized key {}'.format(key))
            if isinstance(value, (list, tuple)):
                values = value
            else:
                values = (value,)
            values = [self._coerce_prop(key, val) for val in values]
            if len(values) == 1:
                values = values[0]
            if key.startswith('@'):
                key = key[1:]
            elif key.startswith(self.CONTEXT):
                key = key[len(self.CONTEXT) + 1:]
            yield key, values

    def part_to_dict(self):
        if self._props['@type'] == as_uri('Object'):
            if len(self._props) == 1:
                return None
            if len(self._props) == 2 and '@id' in self._props:
                return {'@id': self._props['@id']}
        return dict(self._props_to_dict())


class Object(_Base):
    TYPE_PROPERTIES = ('attachment', 'attributed_to', 'audience', 'content',
                       'context', 'name', 'endTime', 'generator', 'icon',
                       'image', 'in_reply_to', 'location', 'preview',
                       'published', 'replies', 'startTime', 'summary', 'tag',
                       'updated', 'url', 'to', 'bto', 'cc', 'bcc',
                       'media_type', 'duration')


class Link(_Base):
    TYPE_PROPERTIES = ('href', 'rel', 'media_type', 'name', 'hreflang',
                       'height', 'width', 'preview')


class Activity(Object):
    TYPE_PROPERTIES = Object.TYPE_PROPERTIES + (
        'actor', 'object', 'target', 'result', 'origin', 'instrument')


class IntransitiveActivity(Activity):
    TYPE_PROPERTIES = Object.TYPE_PROPERTIES + (
        'actor', 'target', 'result', 'origin', 'instrument')


class Collection(Object):
    TYPE_PROPERTIES = Object.TYPE_PROPERTIES + (
        'total_items', 'current', 'first', 'last', 'items')

    def __init__(self, **kwargs):
        items = kwargs.get('items', None)
        if items is not None and 'total_items' not in kwargs:
            if isinstance(items, (list, tuple)):
                count = len(items)
            else:
                count = 1
            kwargs['total_items'] = count
        super().__init__(**kwargs)


class OrderedCollection(Collection):
    pass


class CollectionPage(Collection):
    PAGE_PROPERTIES = ('part_of', 'next', 'prev')
    TYPE_PROPERTIES = Collection.TYPE_PROPERTIES + PAGE_PROPERTIES


class OrderedCollectionPage(CollectionPage, OrderedCollection):
    TYPE_PROPERTIES = (OrderedCollection.TYPE_PROPERTIES
                       + CollectionPage.PAGE_PROPERTIES)


class Accept(Activity):
    pass


class TentativeAccept(Accept):
    pass


class Add(Activity):
    pass


class Arrive(IntransitiveActivity):
    pass


class Create(Activity):
    pass


class Delete(Activity):
    pass


class Follow(Activity):
    pass


class Ignore(Activity):
    pass


class Join(Activity):
    pass


class Leave(Activity):
    pass


class Like(Activity):
    pass


class Offer(Activity):
    pass


class Invite(Offer):
    pass


class Reject(Activity):
    pass


class TentativeReject(Reject):
    pass


class Remove(Activity):
    pass


class Undo(Activity):
    pass


class Update(Activity):
    pass


class View(Activity):
    pass


class Listen(Activity):
    pass


class Read(Activity):
    pass


class Move(Activity):
    pass


class Travel(IntransitiveActivity):
    pass


class Announce(Activity):
    pass


class Block(Ignore):
    pass


class Flag(Activity):
    pass


class Dislike(Activity):
    pass


class Question(IntransitiveActivity):
    pass


# actors


class Application(Object):
    pass


class Group(Object):
    pass


class Organization(Object):
    pass


class Person(Object):
    pass


class Service(Object):
    pass


# Object/Link types

class Relationship(Object):
    TYPE_PROPERTIES = Object.TYPE_PROPERTIES + (
        'subject', 'object', 'relationship')


class Article(Object):
    pass


class Document(Object):
    pass


class Audio(Document):
    pass


class Image(Document):
    pass


class Video(Document):
    pass


class Note(Object):
    pass


class Page(Object):
    pass


class Event(Object):
    pass


class Place(Object):
    TYPE_PROPERTIES = Object.TYPE_PROPERTIES + (
        'accuracy', 'altitude', 'latitude', 'longitude', 'radius', 'units')


class Mention(Link):
    pass


class Profile(Object):
    TYPE_PROPERTIES = Object.TYPE_PROPERTIES + ('describes',)


class Tombstone(Object):
    TYPE_PROPERTIES = Object.TYPE_PROPERTIES + ('former_type', 'deleted')


# properties

def noop(data, _context):
    return data


float_property = noop
string_property = noop
unit_property = noop
lang_property = noop
mime_property = noop
non_neg_int_property = noop
only_object = noop


def duration_property(data, _context):
    return isodate.duration_isoformat(data)


def datetime_property(data, _context):
    return isodate.datetime_isoformat(data)


def any_uri(uri, context):
    if uri.startswith(context):
        uri = uri[len(context):]
        if uri.startswith('#'):
            uri = uri[1:]
    return uri


def link_or_object(data, context):
    if isinstance(data, str):
        return any_uri(data, context)
    if isinstance(data, dict):
        if len(data.keys()) == 1 and '@id' in data.keys():
            return any_uri(data['@id'], context)
        return data
    if isinstance(data, _Base):
        return data
    raise TypeError('Invalid value for link | object {}'.format(data))


link_or_uri = link_or_object
link_or_collection_page = link_or_object
link_or_collection = link_or_object
link_or_image = link_or_object
collection = only_object


def closed_property(data, context):
    if isinstance(data, bool):
        return data
    if isinstance(data, datetime.datetime):
        return datetime_property(data, context)
    return link_or_object(data, context)


PROPERTIES = {
    '@id': any_uri,
    '@type':  any_uri,
    as_uri_prop('actor'): link_or_object,
    as_uri_prop('attachment'): link_or_object,
    as_uri_prop('attributed_to'): link_or_object,
    as_uri_prop('audience'): link_or_object,
    as_uri_prop('bcc'): link_or_object,
    as_uri_prop('bto'): link_or_object,
    as_uri_prop('cc'): link_or_object,
    as_uri_prop('context'): link_or_object,
    as_uri_prop('current'): link_or_collection_page,
    as_uri_prop('first'): link_or_collection_page,
    as_uri_prop('next'): link_or_collection_page,
    as_uri_prop('prev'): link_or_collection_page,
    as_uri_prop('generator'): link_or_object,
    as_uri_prop('icon'): link_or_image,
    as_uri_prop('image'): link_or_image,
    as_uri_prop('in_reply_to'): link_or_object,
    as_uri_prop('instrument'): link_or_object,
    as_uri_prop('last'): link_or_collection_page,
    as_uri_prop('location'): link_or_object,
    as_uri_prop('items'): link_or_object,
    as_uri_prop('one_of'): link_or_object,
    as_uri_prop('any_of'): link_or_object,
    as_uri_prop('closed'): closed_property,
    as_uri_prop('origin'): link_or_object,
    as_uri_prop('object'): link_or_object,
    as_uri_prop('result'): link_or_object,
    as_uri_prop('replies'): collection,
    as_uri_prop('tag'): link_or_object,
    as_uri_prop('target'): link_or_object,
    as_uri_prop('to'): link_or_object,
    as_uri_prop('url'): link_or_uri,
    as_uri_prop('current'): link_or_object,
    as_uri_prop('accuracy'): float_property,
    as_uri_prop('altitude'): float_property,
    as_uri_prop('content'): string_property,
    as_uri_prop('name'): string_property,
    as_uri_prop('duration'): duration_property,
    as_uri_prop('href'): any_uri,
    as_uri_prop('hreflang'): lang_property,
    as_uri_prop('part_of'): link_or_collection,
    as_uri_prop('latitude'): float_property,
    as_uri_prop('longitude'): float_property,
    as_uri_prop('media_type'): mime_property,
    as_uri_prop('start_time'): datetime_property,
    as_uri_prop('end_time'): datetime_property,
    as_uri_prop('published'): datetime_property,
    as_uri_prop('updated'): datetime_property,
    as_uri_prop('deleted'): datetime_property,
    as_uri_prop('radius'): float_property,
    # missing rel
    as_uri_prop('start_index'): non_neg_int_property,
    as_uri_prop('summary'): string_property,
    as_uri_prop('total_items'): non_neg_int_property,
    as_uri_prop('units'): unit_property,
    as_uri_prop('height'): non_neg_int_property,
    as_uri_prop('width'): non_neg_int_property,
    as_uri_prop('subject'): link_or_object,
    as_uri_prop('relationship'): only_object,
    as_uri_prop('describes'): only_object,
    as_uri_prop('former_type'): only_object,
}
